"""
Module listing all public method from the methodo modules

You should put here all methods doing some kind of algorithmic transformation of domain/db models.
"""
from ecodev_sankey.methodo.process_input_file import process_file

__all__ = ['process_file']





"""
Module implementing the base data point version
"""
from typing import Optional

from ecodev_core.sqlmodel_utils import SQLModelWithVal
from sqlalchemy.dialects.postgresql import JSONB
from sqlmodel import Field


class DataPointBase(SQLModelWithVal):
    """
    A datapoint: a name, an optional description and a representation
    """
    name: Optional[str] = Field(index=True)
    description: Optional[str] = Field(index=False)
    nodes_rep: dict =  Field(sa_type=JSONB, nullable=True)




"""
Module impementing a link table between TreeNode and DataPoint
"""
from typing import Optional

from ecodev_core.sqlmodel_utils import SQLModelWithVal
from sqlmodel import Field


class NodeDataPointLink(SQLModelWithVal, table=True):   # type: ignore
    """
    Table storing many to many link between datapoint and tree nodes as
    one datapoint belong to several hierarchies and a node in the hierarchy can refer to multiple
    datapoints

    """
    __tablename__ = 'node_datapoint_link'
    node_id: Optional[int] = Field(default=None,
                                   foreign_key='tree_node.id',
                                   primary_key=True,
                                   index=True
                                   )
    datapoint_id: Optional[int] = Field(default=None,
                                        foreign_key='datapoint.id',
                                        primary_key=True,
                                        index=True
                                        )



"""
Module implementing all tree node retriever methods
"""
from functools import lru_cache
from typing import Iterable
from typing import Optional

import pandas as pd
from ecodev_core import logger_get
from sqlalchemy import exc
from sqlmodel import func
from sqlmodel import select
from sqlmodel import Session

from ecodev_sankey.db_model.tree_node import TreeNode

log = logger_get(__name__)


@lru_cache
def get_tree_node(project_id: int,
                  business_concept: str,
                  level: int,
                  name: str,
                  session: Session
                  ) -> TreeNode:
    """
    Retrieve a tree node given its name
    """
    try:

        return session.exec(select(TreeNode).where(
            TreeNode.project_id == project_id,
            TreeNode.business_concept == business_concept,
            TreeNode.name == name,
            TreeNode.level == level
        )).one()
    except Exception as e:
        if isinstance(e, exc.NoResultFound):
            error = f'{name} was not found in hierarchy {business_concept} for project ' +\
                    f'{project_id} at level {level}'
        else:
            error = f'{type(e).__name__} {name} appears multiple times in hierarchy '\
                f'{business_concept} for project {project_id} at level {level}'
        log.warning(error)

        raise ValueError(error)


def get_row_tree_nodes(project_id: int,
                       row: pd.Series,
                       hierarchies: dict[str, int],
                       session: Session
                       ) -> Iterable[TreeNode]:
    """
    Given a datapoint row, yields its tree nodes.
    Since some hierarchy can have varying depth, we need to iterate backward over h_size to find the
    first non null column for the given hierarchy.
    For actions, some hierarchies can be completely none and need specific handling
    """
    h_start = 0
    for hierarchy, h_size in hierarchies.items():
        try:
            vals = row.iloc[h_start:h_start+h_size].values
            h_start = h_start + h_size
            vals = [handle_number_col(k) for k in vals if not pd.isnull(k)]
            if not vals:
                continue
            name = '_'.join(vals)
            yield get_tree_node(project_id, hierarchy, len(vals)-1, name, session)
        except Exception as e:
            log.critical(f'{type(e).__name__} {e}')
            raise e


def get_node_by_id(node_id: int, session: Session):
    """
    given a node id, return it
    """
    return session.exec(select(TreeNode).where(TreeNode.id == node_id)).one()


def get_all_parent_nodes_ids(node: TreeNode) -> list[int]:
    """
    Given a tree node, return the list of ids of itself and all its parent
    """
    nodes = [node.id]

    if node.parent:
        nodes.extend(get_all_parent_nodes_ids(node.parent))

    return nodes  # type: ignore[return-value]


def get_all_children_nodes_ids(node: TreeNode) -> list[int]:
    """
    Given a tree node, return the list of ids of itself and all its parent
    """
    nodes = [node.id]

    if node.children:
        for child in node.children:
            nodes.extend(get_all_children_nodes_ids(child))

    return nodes  # type: ignore[return-value]


def get_business_axis(project_id: int, session: Session):
    """
    Return the list of tree hierarchies for the given project
    """
    return session.exec(select(TreeNode.business_concept).where
                        (TreeNode.project_id == project_id).distinct()).all()


@lru_cache(maxsize=None)
def get_node_rep(node_id: int, project_id: int, session: Session) -> dict[str, str | None]:
    """
    Given a node_id, return a dict mapping each level of the tree to its name.

    With geography data like this:
    Continent Country
    NA         USA
    NA         Canada
    Europe     France


    it will return {'continent': 'NA', 'Country': 'USA'} when provided with the node for USA
    and {'continent': 'NA', 'Country': None} when provided with the node for NA

    As this is called repeatedly to construct sankey data, we use cache which impose passing
    hashable arguments, thus calling it with node.id then reselecting node from id
    """

    node = session.exec(select(TreeNode).where(TreeNode.id == node_id)).one()

    business_concept = node.business_concept
    if node.parent:
        parent_rep = get_node_rep(node.parent.id, project_id, session)
    else:
        parent_rep = {}
        for k in get_tree_struct(business_concept, project_id, session):
            parent_rep[k] = None
            parent_rep[k+'_id'] = None
    parent_rep[node.level_name] = node.name
    parent_rep[node.level_name+'_id'] = node.id
    return parent_rep.copy()


@lru_cache
def get_tree_struct(business_concept: str,
                    project_id: int,
                    session: Session,
                    ) -> list[str]:
    """
    Given a business concept, return the list of level name:
    'Geography'=> ['Continent', 'Country']
    'LOB'=>['LOB1', 'LOB2', 'LOB3']
    """

    levels = session.exec(select(TreeNode.level,
                                 TreeNode.level_name).distinct().where(
        TreeNode.business_concept == business_concept,
        TreeNode.project_id == project_id
    ).order_by(TreeNode.level)).all()
    return list(zip(*levels))[1]  # type: ignore[return-value]


def get_tree_node_id(project_id: int,
                     business_concept: str,
                     name: str,
                     session: Session) -> int:
    """
    Given a business concept and a name, return a TreeNode id.

    nb: This won't work if several nodes have the same name in different part of the same tree
    """
    return session.exec(select(TreeNode.id)
                        .where((TreeNode.business_concept == business_concept),
                               (TreeNode.name == name),
                               (TreeNode.project_id == project_id))).one()


def get_flat_hierarchy(project_id: int, business_concept: str, session: Session) -> list[str]:
    """
    Given a hierarchy name, return a flat list of choices, covering all levels.
    It looks like this:
    ['Europe',
     'Europe | France'
     'Europe | Germany',
     'Asia',
     'Asia | Japan'
    ]
    """
    choices = []

    if session.exec(select(func.max(TreeNode.level)).where(
            TreeNode.project_id == project_id,
            TreeNode.business_concept == business_concept)).one() == 0:
        return session.exec(select(TreeNode.name).where(
            TreeNode.project_id == project_id,
            TreeNode.business_concept == business_concept
        ).order_by(TreeNode.name)).all()

    top_levels = session.exec(select(TreeNode).where(
        TreeNode.project_id == project_id,
        TreeNode.level == 0,
        TreeNode.business_concept == business_concept)).all()

    for top_level in top_levels:
        choices.extend(_get_level_choices(top_level))

    return sorted(choices)


def _get_level_choices(node: TreeNode,
                       base: Optional[str] = None
                       ) -> list[str]:
    """
    Given a node, return its all hiearchy as a flat list
    """
    choices = []
    name = node.name if not base else f'{base} | {node.name}'
    choices.append(name)
    for child in node.children:
        choices.extend(_get_level_choices(child, name))
    return choices


def get_node_children_ids(base_node_id: int, session: Session) -> list[int]:
    ancestor_nodes = (
        select(TreeNode.id)
        .where(
            (TreeNode.id == base_node_id))
        .cte(name='ancestor_nodes', recursive=True)
    )
    ancestor_nodes = ancestor_nodes.union_all(
        select(TreeNode.id)
        .where(TreeNode.parent_id == ancestor_nodes.c.id)
    )
    all_childrens = session.exec(select(ancestor_nodes)).all()
    return [k for k in all_childrens if k != base_node_id]


def handle_number_col(val: float | str | int) -> str:
    """
    Handle varying type col to return a str
    """
    if isinstance(val, int) or isinstance(val, float):
        val = str(int(val))
    return val.strip()



"""
Module ingesting template to create DataPoint and associated TreeNode
"""
from typing import Any, Callable

import pandas as pd
import progressbar
from ecodev_core import logger_get
from sqlmodel import Session
from sqlmodel.main import SQLModelMetaclass


from ecodev_sankey.constants import DESCRIPTION
from ecodev_sankey.db_model import handle_number_col, get_row_tree_nodes, get_node_rep

log = logger_get(__name__)


def process_datapoints(project_id: int,
                       session: Session,
                       df: pd.DataFrame,
                       hierarchies: dict[str, int],
                       DataPoint: SQLModelMetaclass,
                       row_getters: Callable
                       ) -> None:
    """
    Add all datapoints to the project for all provided years
    """
    try:
        process_datapoints_sheet(df, project_id, hierarchies, session, DataPoint, row_getters)
    except Exception as e:
        log.critical(f'{type(e).__name__} {e}')
        raise ValueError(f'Datapoint insertion: {type(e).__name__} {e}')


def process_datapoints_sheet(datapoint_df: pd.DataFrame,
                             project_id: int,
                             hierarchies: dict[str, int],
                             session: Session,
                             DataPoint: SQLModelMetaclass,
                             row_getters: Callable) -> None:
    """
    Given a footprint sheet, create all datapoints.
    For old years, we ignore targets as those datapoints are only provided for historical
    comparisons and not affected by targets
    """

    for _, datapoint_row in progressbar.progressbar(datapoint_df.iterrows(),
                                                    max_value=len(datapoint_df),
                                                    redirect_stdout=False):

        name = '_'.join(handle_number_col(k) for k
                        in datapoint_row.iloc[:sum(hierarchies.values())].values
                        if not pd.isna(k))
        tree_nodes = list(get_row_tree_nodes(project_id, datapoint_row, hierarchies, session))
        nodes_rep: dict[str, Any] = {}
        for node in tree_nodes:
            nodes_rep |= get_node_rep(node.id, project_id, session)
        session.add(DataPoint(name=name, description =datapoint_row.get(DESCRIPTION),
                               project_id=project_id,
                               nodes=tree_nodes,
                               nodes_rep=nodes_rep,
                              **row_getters(datapoint_row)
                               ))

    session.commit()



"""
Module implementing all datatpoint retriever methods
"""
from typing import Any, Iterator, Callable

from sqlalchemy import func
from sqlalchemy import tuple_
from sqlmodel import Session
from sqlmodel import select
from sqlmodel.main import SQLModelMetaclass

from ecodev_sankey.constants import ID
from ecodev_sankey.db_model.empty_datapoint_list_error import EmptyDatapointList
from ecodev_sankey.db_model.node_datapoint_link import NodeDataPointLink
from ecodev_sankey.db_model.retrievers.tree_node_retrievers import get_node_rep
from ecodev_sankey.db_model.tree_node import TreeNode


def get_sankey_datapoints(project_id: int,
                    filters: dict[str, list[int]],
                    field_adders: Callable,
                    session: Session,
                    DataPoint: SQLModelMetaclass
                    ) -> Iterator[dict]:
    """
    Returns a DataFrame to be used for generating the Sankey diagram.
    The data fetch is in two steps:
    - 1) retrieve the required data from (data_point, geography, entity, category) dbs
    - 2) convert each data to a DataFrame records.
    """
    for datapoint in retrieve_datapoints_with_filters(project_id, filters, session, DataPoint):
        nodes_rep: dict[str, Any] = {}
        for node_id in session.exec(select(NodeDataPointLink.node_id).where(
                NodeDataPointLink.datapoint_id == datapoint.id)):
            nodes_rep |= get_node_rep(node_id, project_id, session)
        yield {ID: datapoint.id} | field_adders(datapoint, session) | nodes_rep



def retrieve_datapoints_from_hierarchy_node(project_id: int,
                                            base_node_id: int,
                                            session: Session,
                                            DataPoint: SQLModelMetaclass
                                            ) -> list[SQLModelMetaclass]:
    """
    Return all datapoints where one of their node is in the subtree starting with base_node_id
    """
    ancestor_nodes = (
        select(TreeNode.id)
        .where(
            (TreeNode.id == base_node_id))
        .cte(name='ancestor_nodes', recursive=True)
    )
    ancestor_nodes = ancestor_nodes.union_all(
        select(TreeNode.id)
        .where(TreeNode.parent_id == ancestor_nodes.c.id)
    )

    stmt = (
        select(DataPoint)
        .join(NodeDataPointLink, DataPoint.id == NodeDataPointLink.datapoint_id)
        .join(TreeNode, NodeDataPointLink.node_id == TreeNode.id)
        .where(TreeNode.id.in_(select(ancestor_nodes.c.id)),  # type: ignore[union-attr]
               DataPoint.project_id == project_id
               )
        .distinct()
    )
    return session.exec(stmt).all()


def retrieve_datapoints_from_multi_tree(project_id: int,
                                        base_node_ids: list[int],
                                        session: Session,
                                        DataPoint: SQLModelMetaclass
                                        ) -> list[int]:
    """
    Given a list of tree node, return the list of datapoint that belong to all trees
    """
    eligible_per_node = [set([point.id
                              for point in retrieve_datapoints_from_hierarchy_node(project_id,
                                                                                   base_node_id,
                                                                                   session,
                                                                                   DataPoint)])
                         for base_node_id in base_node_ids]
    if not eligible_per_node:
        raise EmptyDatapointList('no datapoints for provided list')
    return list(set.intersection(*eligible_per_node))  # type: ignore[arg-type]


def retrieve_datapoint(datapoint_id: int,
                       session: Session,
                       DataPoint: SQLModelMetaclass
                       ) -> SQLModelMetaclass:
    """
    Retrieve a single datapoint by id
    """
    return session.exec(select(DataPoint).where(DataPoint.id == datapoint_id)).one()


def retrieve_datapoints(datapoint_ids: list[int], session: Session) -> list[SQLModelMetaclass]:
    """
    Retrieve a list of datapoints  by id
    """
    return session.exec(select(DataPoint).where(DataPoint.id.in_(datapoint_ids))).all()  # type: ignore[union-attr] # noqa: E501


def retrieve_datapoints_with_filters(
    project_id: int,
    hierarchy_filters: dict[str, list[int]],
    session: Session,
    DataPoint: SQLModelMetaclass
) -> list[SQLModelMetaclass]:
    """
    Given a hierarchy_filters giving a list of allowed node for multiple hierarchies,
    return the list of datapoint that satisfy all hierarchies where satisfying a hierarchy
    mean the datapoint node in this hierarchy belong to any subtrees starting from the provided
    node ids.
    With {'Geography' : [1,2],'Activity['3']}
    a datapoint is eligible if his hierarchy node is under 1 or 2  AND its activity node under 3

    To do so, we build a CTE with all base_node_id and their children, each tagged with their
    business concept, join with datapoint and group, keeping datapoints that match on
    all business_concept
    """

    if not hierarchy_filters:
        return session.exec(select(DataPoint).where(DataPoint.project_id == project_id)).all()

    # Flatten to list of (business_concept, base_id) tuples
    concept_base_pairs = [
        (bc, base_id)
        for bc, base_ids in hierarchy_filters.items()
        for base_id in base_ids
    ]

    # Anchor: all base nodes for all (business_concept, id) pairs
    anchor = (
        select(
            TreeNode.id.label('node_id'),  # type: ignore[union-attr]
            TreeNode.business_concept.label(  # type: ignore[attr-defined]
                'business_concept'),  # type: ignore[union-attr]
        )
        .where(tuple_(TreeNode.business_concept,
                      TreeNode.id).in_(concept_base_pairs))
    )

    # Recursive: children of nodes in the CTE, keeping the business_concept
    cte = anchor.cte('all_subtree_nodes', recursive=True)
    recursive = (
        select(
            TreeNode.id.label('node_id'),  # type: ignore[union-attr]
            cte.c.business_concept,
        )
        .where(TreeNode.parent_id == cte.c.node_id)
        .where(TreeNode.business_concept == cte.c.business_concept)
    )
    all_nodes_cte = cte.union_all(recursive)

    # Join DataPoints to this CTE
    stmt = (
        select(DataPoint)
        .join(NodeDataPointLink, DataPoint.id == NodeDataPointLink.datapoint_id)
        .join(TreeNode, NodeDataPointLink.node_id == TreeNode.id)
        .join(all_nodes_cte, TreeNode.id == all_nodes_cte.c.node_id)
        .where(
            DataPoint.project_id == project_id,
        )
        .group_by(DataPoint.id)
        .having(
            func.count(func.distinct(all_nodes_cte.c.business_concept)) == len(hierarchy_filters)
        )

    )

    return session.exec(stmt).all()



"""
Module implementing template processing methods
"""
import pandas as pd
from ecodev_core import logger_get
from sqlalchemy import delete
from sqlmodel import Session
from sqlmodel.main import SQLModelMetaclass

from ecodev_sankey.db_model import NodeDataPointLink
from ecodev_sankey.db_model import TreeNode
from ecodev_sankey.methodo.datapoint_processing import process_datapoints
from ecodev_sankey.methodo.hierarchy_processing import create_tree_hierarchies

log = logger_get(__name__)


def process_file(df: pd.DataFrame,
                 project_id: int,
                 hierarchies: dict[str, int],
                 DataPoint: SQLModelMetaclass,
                 row_getters: dict,
                 session: Session) -> None:
    """
    Process an Excel file to import project data including hierarchies and datapoints.

    Args:
        token: Authentication token for user validation
        xl: Excel file containing project data
        project_id: ID of the project to import data for
        session: Database session for operations
    """
    log.warning(f'importing file for project {project_id}')

    log.info(f'delete {project_id}')
    delete_datapoints(project_id, session, DataPoint)
    log.warning('importing hierarchies')
    create_tree_hierarchies(project_id, session, df, hierarchies)
    log.warning('importing datapoints')
    process_datapoints(project_id, session, df, hierarchies, DataPoint, row_getters)


def delete_datapoints(project_id: int, session: Session, DataPoint: SQLModelMetaclass) -> None:
    """
    Delete all existing data for a project before importing new data.

    Args:
        project_id: ID of the project to delete data for
        session: Database session for operations
    """

    log.info(f'deleting {NodeDataPointLink}')
    session.exec(delete(NodeDataPointLink).where(NodeDataPointLink.node_id == TreeNode.id,
                                                 TreeNode.project_id == project_id))
    for table in [DataPoint, TreeNode]:
        log.info(f'deleting {table}')
        # type: ignore[attr-defined]
        session.exec(delete(table).where(table.project_id == project_id))
    session.commit()



"""
Module listing all public sankey methods
"""
from ecodev_sankey.db_model import TreeNode, NodeDataPointLink
from ecodev_sankey.db_model.datapoint_base import DataPointBase
from ecodev_sankey.db_model.retrievers.datapoint_retrievers import get_sankey_datapoints
from ecodev_sankey.methodo.process_input_file import delete_datapoints
from ecodev_sankey.db_model.retrievers.datapoint_retrievers import retrieve_datapoints_with_filters
from ecodev_sankey.db_model.retrievers.tree_node_retrievers import get_node_by_id
from ecodev_sankey.db_model.retrievers.tree_node_retrievers import get_flat_hierarchy, \
    get_business_axis, get_tree_struct
from ecodev_sankey.methodo.compute_sankey_data import create_graph_from_columns
from ecodev_sankey.methodo.filter_helpers import get_filter_dict
from ecodev_sankey.methodo.filter_helpers import get_merged_filter
__all__ = ['DataPointBase',
           'TreeNode',
           'NodeDataPointLink',
           'get_sankey_datapoints',
           'delete_datapoints',
           'retrieve_datapoints_with_filters',
           'get_node_by_id',
           'get_flat_hierarchy',
           'get_business_axis',
           'get_tree_struct',
           'get_merged_filter',
           'create_graph_from_columns',
           'get_filter_dict'
           ]


"""
Module implementing the tree node class
"""
from typing import Optional

from ecodev_core.sqlmodel_utils import SQLModelWithVal
from sqlalchemy import Index
from sqlmodel import Field
from sqlmodel import Relationship


class TreeNodeBase(SQLModelWithVal):
    """
    Base table for the node hierarchies
    Business_concept: name of the hierarchy
    name: name of the node
    level: depth of the node in the hierarchy
    """
    business_concept: str = Field(index=True)
    name: str = Field(index=True)
    level: int = Field(index=True)
    level_name: str = Field(index=True)


class TreeNode(TreeNodeBase, table=True): # type: ignore
    """
    db version of TreeNodeBase, in many-to-many relationship with a project specific DataPoint class
    """
    __tablename__ = 'tree_node'
    id: Optional[int] = Field(default=None, primary_key=True)
    parent_id: Optional[int] = Field(default=None, foreign_key='tree_node.id', index=True)
    parent: Optional['TreeNode'] = Relationship(
        back_populates='children',
        sa_relationship_kwargs=dict(
            remote_side='TreeNode.id'  # notice the uppercase to refer to this table class
        )
    )
    children: list['TreeNode'] = Relationship(back_populates='parent',
                                              cascade_delete=True)
    project_id: Optional[int] = Field(default=None, foreign_key='project.id', index=True)
    def __repr__(self) -> str:
        return f'TreeNode: {self.name}({self.business_concept}[{self.level}]) {self.id}'

    __table_args__ = (
        Index('idx_tree_node_composite', 'project_id', 'level', 'business_concept'),
    )



"""
Module listing all sankey related constants
"""
LABEL = 'label'
NODE_ID = 'node-id'
COLOR = 'color'
ROOT = 'root'
VALUE = 'value'
NUM_DATAPOINTS = 'num_datapoints'
NUM_CHILDREN = 'num-children'
ID = 'id'
DESCRIPTION = 'description'
HIERARCHIES_SHEET = 'Hierarchies'
HIERARCHY = 'hierarchy'
DEPTH = 'depth'


"""
Module implementing a message specific error for empty datapoint associated data
"""
class EmptyDatapointList(Exception):
    """
    message specific error for empty datapoint associated data
    """
    def __init__(self, message):
        self.message = message



"""
Module implementing al filtering methods
"""
from collections import defaultdict

from sqlmodel import Session

from ecodev_sankey.constants import ID
from ecodev_sankey.db_model.retrievers.tree_node_retrievers import get_node_children_ids, \
    get_tree_node, get_node_by_id


def get_filter_dict(project_id: int,
                    hierarchies: list[str],
                    filter_values: list[list[str]],
                    session: Session
                    ) -> dict[str, list[int]]:
    """
    Given a list of filter values for each hierarchy, possibly overlapping,
    return a dict mapping hierarchy to a list of mutually exclusive node_ids matching the filter.

    To do so we merge overlapping filters by keeping only parent if parent and children are
    simultaneously present

    """
    filter_dict = {}
    for hierarchy, h_filter_values in zip(hierarchies, filter_values):

        if h_filter_values:
            reduced_filters = _get_minimal_filter_list(h_filter_values)
            filter_dict[hierarchy] = _get_filter_nodes(
                project_id, hierarchy, reduced_filters, session)
    return filter_dict


def _get_minimal_filter_list(filters: list[str]) -> dict[int, list[str]]:
    """
    Given a list of filter values at different levels of the hierarchy,
    * remove specifics filters covered by a parent one
    * return a dict mapping each hierarchy level to the list of filters kept

    ['A | B', 'A', 'C | D', 'E | F | G']
    =>
    {0:['A'],1:['D'], 2:['G']

    }
    """
    reduced_filters = defaultdict(list)
    normalized = [k.replace(' | ', '|') for k in filters]
    kept = []
    for label in sorted(normalized, key=lambda x: x.count('|')):
        levels = label.split('|')
        parents = ['|'.join(levels[:i]) for i in range(1, len(levels)+1)]
        if not any(parent in kept for parent in parents):
            kept.append(label)
    for k in kept:
        reduced_filters[k.count('|')].append(k.split('|')[-1])
    return reduced_filters


def _get_filter_nodes(project_id: int,
                      business_concept: str,
                      reduced_filters: dict[int, list[str]],
                      session: Session
                      ) -> list[int]:
    """
    Given the dict provided by _get_minimal_filter_list, return a dict with only node ids as values
    for storing in dcc.Store
    """
    return [get_tree_node(project_id, business_concept, level, name, session).id
            for level, names in reduced_filters.items()
            for name in names]


def merge_filter_and_node(node, filters, session):
    """
    Given a list of filters, add the selected node constraint. It implies keeping the intersection
    of the constraint from the filters and the node

    For the business concept of the node there are two cases:
    The node is a parent of some of the filters, in which case we keep all children and remove other
    filters. Otherwise, the node is a child of some filter (or there are no filter for the business
    concept) in which case we keep the node.


    """
    children = get_node_children_ids(node.id, session)
    filters[node.business_concept] = [k for k in filters.get(node.business_concept, [])
                                      if k in children] or [node.id]
    return filters

def get_merged_filter(project_id: int,
                      node_id: int,
                      selectors_ids: list[str],
                      selector_values: list[list[str]],
                      session: Session):
    """
    Forge a filter out of the passed selector information, then keeps the intersection of
     the constraint from the filters and the node

    For the business concept of the node there are two cases:
    The node is a parent of some of the filters, in which case we keep all children and remove other
    filters. Otherwise, the node is a child of some filter (or there are no filter for the business
    concept) in which case we keep the node.
    """
    filters = get_filter_dict(project_id, [selector_id[ID] for selector_id in selectors_ids],
                              selector_values, session)
    return merge_filter_and_node(get_node_by_id(node_id, session), filters, session)


"""
Module converting a list of data points (at df format to ease the conversion)
 into a networkx DiGraph. This corresponds to a Directed Acyclic Graph (or DAG). networkx provides
 very convenient helper methods to do depth-first searches, shortest path distance computations...
"""
import networkx as nx
import pandas as pd
from networkx.classes.digraph import DiGraph

from ecodev_sankey.constants import COLOR, VALUE
from ecodev_sankey.constants import LABEL
from ecodev_sankey.constants import NODE_ID
from ecodev_sankey.constants import NUM_CHILDREN
from ecodev_sankey.constants import NUM_DATAPOINTS
from ecodev_sankey.constants import ROOT


def create_graph_from_columns(df: pd.DataFrame,
                              cols: list[str],
                              names : list[str],
                              colors: list[str]) -> DiGraph:
    """
    Create a graph out of the passed dataframe, given a list of cols

    NB: We create a root node to ease the DAG generation
    """
    graph = nx.DiGraph()
    graph.add_nodes_from([(0, {COLOR: colors[0], VALUE: 100, LABEL: ROOT})])
    for depth, column in enumerate(cols):
        _add_nodes_from_column(graph, df, column, depth, names, colors)
    _add_edges_from_columns(graph, df, cols, names, colors)
    _add_colors(graph)
    _add_num_children(graph)
    return graph


def get_node_labels(graph: DiGraph) -> list[dict[str, str]]:
    """
    Get all node labels for sankey node selector
    """
    labels = set([lab for x in graph.nodes if (lab := graph.nodes[x][LABEL]) != ROOT
                  and not lab.endswith(' - Other')])
    return [{LABEL: lab, VALUE: lab} for lab in labels]


def _add_nodes_from_column(graph: DiGraph,
                           raw_df: pd.DataFrame,
                           col: str, depth: int,
                           names: list[str],
                           colors: list[str]
                           ) -> None:
    """
    Add nodes to the passed graph corresponding to column in df.

    NB: subtlety for the higher level depth: in that case we want to create edges between those
       high level nodes and the dag root node.

    NB2: here we do sort in ascending order the nodes wrt to the quantitative column,
     as we color them in this order (and we would rather prefer to have the nodes with the
     highest value wrt to the quantitative column to color their children than the other way around
    """
    df = raw_df[[col, col+'_id', *names]].groupby(col).agg(
        {col+'_id': 'first'} |  {x: 'sum' for x in names}
    ).reset_index()
    for jdx, x in df.sort_values(names).iterrows():

        node = _infos(raw_df[raw_df[col] == x[col]], jdx, x,  x[col], x[col+'_id'], names, colors)
        graph.add_nodes_from([(len(graph.nodes), node)])
        if depth == 0:
            edge = {COLOR: colors[0],  LABEL: f'source: root, target: {x[col]}'} | {
                field: x.loc[field] for field in names}
            graph.add_edges_from([(0, len(graph.nodes) - 1, edge)])


def _add_edges_from_columns(graph: DiGraph,
                            df: pd.DataFrame,
                            cols: list[str],
                            names: list,
                            colors: list[str]) -> None:
    """
    Add all edges (not related to root) to the passed graph.
    """
    nodes = {graph.nodes[node][LABEL]: node for node in graph.nodes}
    for first_col, second_col in zip(cols, cols[1:]):
        _add_edges_from_column_pairs(graph, df, first_col, second_col, nodes, names, colors)


def _add_edges_from_column_pairs(
        graph: DiGraph,
        df: pd.DataFrame,
        f_col: str,
        s_col: str,
        nodes: dict[str, int],
        names: list[str],
        colors: list[str]
) -> None:
    """
    Add all edges between nodes of f_col and s_col in the flatten df data.

    NB: Beware, huge subtlety, we forbid here cycles. To be discussed further

    NB2: Beware, huge subtlety number 2: for node who have SEVERAL (not 1, several) children and
    one of them is None, we replace this None with the following node "<node name> - Other".
    """
    dg = df[[f_col, s_col, *names]].groupby([f_col, s_col], dropna=False)[names].sum().reset_index()
    for idx, x in dg.sort_values(names, ascending=False).iterrows():
        edge = {COLOR: colors[0], LABEL: f'source: {x[f_col]} target: {x[s_col]}'} | {
            name: x.loc[name] for name in names}
        graph.add_edges_from([(nodes[x[f_col]], nodes[x[s_col]], edge)])


def _infos(df: pd.DataFrame,
           idx: int,
           x: pd.Series,
           label: str,
           node_id: int,
           names: list[str],
           colors: list[str]) -> dict:
    """
    Get node information. If the EF units are not the same for all data point under this node, only
    returns emission and number of data points under the node. If the EF unit is indeed unique,
    add the number of distinct EF, the unit, the weighted EF mean (weighted by emissions) and std.
    """
    return {
        NUM_DATAPOINTS: len(df),
        LABEL: label,
        COLOR: colors[idx % len(colors)],
        NODE_ID: node_id,
    } | {name: x[name] for name in names}


def _add_colors(graph: DiGraph) -> None:
    """
    Add all edges colors for all edges not related to root.
    """
    for node in [x for x in nx.dfs_preorder_nodes(graph, source=0, depth_limit=1) if x != 0]:
        for edge in [x for x in nx.dfs_edges(graph, source=node)]:
            graph.edges[edge][COLOR] = graph.nodes[node][COLOR]


def _add_num_children(graph: DiGraph) -> None:
    """
    Add the number of children as a node attribute
    """
    for node in graph.nodes:
        graph.nodes[node][NUM_CHILDREN] = len(list(nx.dfs_preorder_nodes(graph, source=node)))



"""
Module implementing all hierarchy processing methods
"""
from typing import Optional

import pandas as pd
from sqlmodel import Session

from ecodev_sankey.constants import HIERARCHY, DEPTH, HIERARCHIES_SHEET
from ecodev_sankey.db_model import TreeNode, handle_number_col


def get_hierarchies(xl: pd.ExcelFile) -> dict[str, int]:
    """
    Parse the hierarchy sheet to get depth per tree
    """

    return {row[HIERARCHY]: row[DEPTH] for _, row in xl.parse(HIERARCHIES_SHEET).iterrows()}


def create_level(project_id: int,
                 hierarchy_df: pd.DataFrame,
                 business_concept: str,
                 session: Session,
                 parent: Optional[TreeNode] = None,
                 level: int = 0
                 ) -> None:
    """
    Recursively create the tree based on a dataframe.
    The dataframe is expected to look like this
    depth1|depth2|depth3
    Europe|France| Paris
    Europe|France| Toulouse
    Europe| Spain| Madrid
    NA    | USA  | Portland
    NA    | USA  | Salem

    """
    for name, sub_df in hierarchy_df.groupby(hierarchy_df.columns[0]):
        sub_df = sub_df[sub_df.columns[1:]]

        name = handle_number_col(name)
        if parent:
            name = f'{parent.name}_{name}'
        node = TreeNode(business_concept=business_concept,
                        name=name,
                        level=level,
                        level_name=hierarchy_df.columns[0],
                        project_id=project_id,
                        parent=parent
                        )
        session.add(node)
        if sub_df.shape[1] > 0:
            # handle subtree with only None
            create_level(project_id, sub_df, business_concept, session, node, level+1)
        node.parent = parent
    session.commit()


def create_tree_hierarchies(project_id: int,
                            session: Session,
                            data_df: pd.DataFrame,
                            hierarchies: dict[str, int]
                            ) -> None:
    """
    Parse the xl to create the hierarchies.
    """
    col_idx = 0
    for hierarchy, h_size in hierarchies.items():
        hierarchy_df = data_df[data_df.columns[col_idx:col_idx+h_size]].drop_duplicates()
        col_idx += h_size
        create_level(project_id, hierarchy_df, hierarchy, session)



"""
Module listing all db methods classes used in methodo
"""
from ecodev_sankey.db_model.node_datapoint_link import NodeDataPointLink
from ecodev_sankey.db_model.tree_node import TreeNode
from ecodev_sankey.db_model.retrievers.tree_node_retrievers import handle_number_col, \
    get_row_tree_nodes, get_node_rep
__all__ = ['NodeDataPointLink', 'TreeNode', 'handle_number_col', 'get_row_tree_nodes',
           'get_node_rep']